#include "circuit.h"

#include "assert.h"

using namespace atpg_ls;

int Gate::cal_propagate_len(bool x) {

    int fpl[2];
    fpl[0] = fpl[1] = 0;
    if(isPO) {
        fpl[!value] = 0;
        fpl[value] = 0;
        return fpl[x];
    }

    std::queue<Gate*> q;

    for(Gate* out : fanouts) {
        if(!out->is_detected(this)) continue;
        fpl[!value] = std::max(fpl[!value], out->fault_propagated_len[!out->value] + 1);
    }
    
    return fpl[x];
}

bool Gate::is_detected(Gate* one_of_input) {
    one_of_input->value = !one_of_input->value;
    bool detect = (cal_value() != value);
    one_of_input->value = !one_of_input->value;
    return (cal_value() == value) && detect;
}

bool Gate::is_propagated() {
    return fault_detected[0] || fault_detected[1];
}

bool Gate::cal_fault_detected(bool x) {
    if(isPO) {
        if(x == 0) return value;
        else return !value;
    }

    bool sa0 = 0;
    bool sa1 = 0;

    for(Gate* out : fanouts) {
        if(!out->is_propagated()) continue;

        if(out->cal_value() != out->value) continue;

        this->value = !this->value;
        bool detect = out->cal_value() != out->value;
        this->value = !this->value;
        if(!detect) continue;

        sa0 |= this->value;
        sa1 |= !this->value;
    }
    if(x == 0) return sa0;
    else return sa1;
}

int Gate::cal_value() {
    int res;

    switch(type) {
        case NOT:
            res = !fanins[0]->value;
            break;
        case LINE:
            // pass through
        case BUF:
            res = fanins[0]->value;
            break;
        case AND:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res &= fanins[i]->value;
            }
            break;
        case NAND:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res &= fanins[i]->value;
            }
            res = !res;
            break;
        case OR:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res |= fanins[i]->value;
            }
            break;
        case NOR:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res |= fanins[i]->value;
            }
            res = !res;
            break;
        case XOR:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res ^= fanins[i]->value;
            }
            break;
        case XNOR:
            res = fanins[0]->value;
            for(int i=1; i<fanins.size(); i++) {
                res ^= fanins[i]->value;
            }
            res = !res;
            break;
        case INPUT:
            res = value;
            break;
        default:
            assert(false);
            break;
    }
    return res;
}

int Gate::lut_get_value() {
    return parent_lut->fault_table[id_in_lut][parent_lut->input_var<<1|parent_lut->value].value;
}


void Gate::lut_get_fault(int *fd, int *fpl) {
    LUT::FaultInfo &info = parent_lut->fault_table[id_in_lut][parent_lut->input_var<<1|parent_lut->value];
    fd[0] = info.fd[0] && parent_lut->fd[!parent_lut->value];
    fd[1] = info.fd[1] && parent_lut->fd[!parent_lut->value];
    fpl[0] = info.fpl[0] + info.fd[0] * parent_lut->fpl[!parent_lut->value];
    fpl[1] = info.fpl[1] + info.fd[1] * parent_lut->fpl[!parent_lut->value];
}